/*
	ncbricks - ncurses based brick breaker
	Copyright (C) 2017-2018  Aron Barath

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
	02110-1301  USA
*/

#include "ncbricks.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <assert.h>

WINDOW* g_stat_win;
WINDOW* g_game_win;
struct paddle g_paddle;
struct ptrarr g_balls;
struct ptrarr g_bricks;
struct ptrarr g_charms;
struct ptrarr g_missiles;
bool g_use_colors = TRUE;
int g_board_top;
int g_board_left;
int g_board_width;
int g_board_height;
int g_paddle_y;
size_t g_num_unbreakable;
int g_level;
unsigned int g_life;
unsigned int g_ammo;
unsigned long long g_score;
int g_backdef;
int g_atomicball;
int g_slowball;

static int start_balls = 3;
static int start_ammo = INITIAL_AMMO;
static bool need_status_update = FALSE;
static bool keep_going = TRUE;
static bool board_invalid = FALSE;
static bool wait_ball_launch = FALSE;
static unsigned int atomic_time = 0;
static unsigned int slow_time = 0;

void invalidate_status()
{
	need_status_update = TRUE;
}

void invalidate_board()
{
	board_invalid = TRUE;
}

void add_score(unsigned int n)
{
	g_score += n;
	need_status_update = TRUE;
}

void add_ammo(unsigned int n)
{
	g_ammo += n;
	need_status_update = TRUE;
}

void add_life(int n)
{
	g_life += n;
	need_status_update = TRUE;
}

void enable_atomic_ball()
{
	g_atomicball = 1;
	ball_hide_all();
	atomic_time = 10 * BALL_ATOMICTIME;
	ball_show_all();
}

void enable_slow_ball()
{
	g_slowball = 1;
	slow_time = 10 * BALL_SLOWTIME;
}

static void clear_board()
{
	wclear(g_game_win);
	wbkgd(g_game_win, ' ');
	wrefresh(g_game_win);
	wborder(g_game_win, 0, 0, ' ', ' ', ACS_VLINE, ACS_VLINE, 0, 0);
	redrawwin(g_game_win);
}

static void draw_board()
{
	clear_board();

	/* TODO: draw stuff */

	wrefresh(g_game_win);
}

static void update_status()
{
	int i;
	int part = GAME_WIDTH / 5;

	wmove(g_stat_win, 0, 0);
	for(i=0;i<GAME_WIDTH;++i)
	{
		waddch(g_stat_win, '_');
	}

	if(g_life>MAX_LIFE) { g_life = MAX_LIFE; }
	if(g_ammo>MAX_AMMO) { g_ammo = MAX_AMMO; }

	mvwprintw(g_stat_win, 0, 1+0*part, "Level:_%i", g_level);
	mvwprintw(g_stat_win, 0, 1+1*part, "Balls:_%i", g_life);
	mvwprintw(g_stat_win, 0, 1+2*part, "Ammo:_%i", g_ammo);
	mvwprintw(g_stat_win, 0, 1+3*part, "Score:_%llu", g_score);

	wrefresh(g_stat_win);
	need_status_update = FALSE;
}

void show_message(const char* const msg)
{
	const size_t len = strlen(msg);
	WINDOW* const win = newwin(1, len, GAME_HEIGHT/2, (GAME_WIDTH-len)/2);
	assert(NULL!=win);
	wattron(win, COLOR_PAIR(CPAIR_TEXT));
	mvwprintw(win, 0, 0, "%s", msg);
	touchwin(win);
	redrawwin(win);
	wrefresh(win);
	keypad(win, TRUE);
	cbreak();
	wgetch(win);
	delwin(win);
	redrawwin(g_game_win);
	wrefresh(g_game_win);
	keypad(g_game_win, TRUE);
	halfdelay(1);
}

static void welcome_screen()
{
}

static void new_ball()
{
	g_atomicball = 0;
	g_slowball = 0;

	paddle_hide(&g_paddle);
	paddle_reset(&g_paddle);
	paddle_show(&g_paddle);
	ball_spawn();
	ball_show_all();
	wait_ball_launch = TRUE;

	--g_life;
	need_status_update = TRUE;
}

static void new_game()
{
	bool level_load_result;

	ball_delete_all();
	brick_delete_all();
	charm_delete_all();
	missile_delete_all();
	clear_board();

	g_level = 1;
	g_score = 0;
	g_ammo = start_ammo;
	g_life = start_balls;
	g_backdef = 0;
	g_atomicball = 0;
	g_slowball = 0;

	new_ball();
	level_load_result = brick_load_data(g_level);

	assert(level_load_result);

	need_status_update = TRUE;
	invalidate_board();
}

static bool next_level()
{
	/* earn some points */

	g_score += SCORE_LEVEL;

	/* set next level */

	++g_level;
	need_status_update = TRUE;

	/* clear the board (we may have multiple balls!) */

	ball_delete_all();
	brick_delete_all();
	charm_delete_all();
	missile_delete_all();
	clear_board();

	/* spawn a new ball with paddle reset */

	++g_life;
	new_ball();

	/* load bricks */

	invalidate_board();
	return brick_load_data(g_level);
}

static void shoot_missile()
{
	if(0==g_ammo) { return; }
	--g_ammo;
	need_status_update = TRUE;

	missile_spawn(&g_paddle);
}

static void game_loop()
{
	int tenths = 0;
	int parity = 0;

	for(;;)
	{

		if(!wait_ball_launch && 10==(++tenths))
		{
			add_score(2);
			tenths = 0;
		}

		if(0<atomic_time && 0==(--atomic_time))
		{
			ball_hide_all();
			g_atomicball = 0;
			ball_show_all();
		}

		if(0<slow_time)
		{
			parity ^= 1;

			if(0==(--slow_time))
			{
				g_slowball = 0;
			}
		}

		if(need_status_update)
		{
			update_status();
		}

		if(board_invalid)
		{
			wrefresh(g_game_win);
			board_invalid = FALSE;
		}

		const int key = wgetch(g_game_win);

		switch(key)
		{
		case 's':
		case KEY_LEFT:
			if(!wait_ball_launch) { paddle_move(&g_paddle, -PADDLE_SMALL); }
			break;

		case 'k':
		case KEY_RIGHT:
			if(!wait_ball_launch) { paddle_move(&g_paddle, +PADDLE_SMALL); }
			break;

		case 'a':
			if(!wait_ball_launch) { paddle_move(&g_paddle, -PADDLE_LARGE); }
			break;

		case 'l':
			if(!wait_ball_launch) { paddle_move(&g_paddle, +PADDLE_LARGE); }
			break;

		case 'd':
			if(!wait_ball_launch) { paddle_move(&g_paddle, -PADDLE_TINY); }
			break;

		case 'j':
			if(!wait_ball_launch) { paddle_move(&g_paddle, +PADDLE_TINY); }
			break;

		case ' ':
			if(!wait_ball_launch)
			{
				shoot_missile();
			}
			else
			{
				wait_ball_launch = FALSE;
			}
			break;

		case 'p':
			show_message("Game paused. Press any key to continue.");
			break;

		case 'q':
			keep_going = FALSE;
			return;
		}

		animate_all(!wait_ball_launch && (!g_slowball || parity));

		if(0==g_balls.size)
		{
			if(0==g_life)
			{
				show_message("Game over!");
				return;
			}

			new_ball();
		}
		else
		if(g_num_unbreakable==g_bricks.size)
		{
			update_status();
			wrefresh(g_game_win);
			show_message("Level cleared. Good work!");
			if(!next_level())
			{
				// TODO: leaderboard
				return;
			}
		}
	}
}

int main(int argc, char* argv[])
{
	srand((unsigned int)time(NULL));

	initscr();

	if(COLS<GAME_WIDTH || LINES<GAME_HEIGHT)
	{
		endwin();
		fprintf(stderr, "You need a terminal with a resolution of at least %ix%i.\n", GAME_HEIGHT, GAME_WIDTH);
		return 1;
	}

	if(!has_colors())
	{
		g_use_colors = FALSE;
	}

	noecho();
	curs_set(0);

	if(NULL==(g_stat_win=newwin(1, GAME_WIDTH, 0, 0)))
	{
		endwin();
		fprintf(stderr, "Failed to create status bar window.\n");
		return 1;
	}

	if(NULL==(g_game_win=newwin(GAME_HEIGHT-1, GAME_WIDTH, 1, 0)))
	{
		endwin();
		fprintf(stderr, "Failed to create game window.\n");
		return 1;
	}

	g_board_top    = 0;
	g_board_left   = 1; /* because of status bar */
	g_board_width  = GAME_WIDTH - 2; /* because of the border */
	g_board_height = GAME_HEIGHT - 1; /* because of status bar */
	g_paddle_y     = g_board_height - 1;

	if(g_use_colors)
	{
		const short brick_bg = COLOR_BLACK;

		start_color();
		init_pair(CPAIR_TEXT,    COLOR_CYAN,    COLOR_BLACK);
		init_pair(CPAIR_BALL,    COLOR_WHITE,   COLOR_BLACK);
		init_pair(CPAIR_BALL2,   COLOR_RED,     COLOR_BLACK);
		init_pair(CPAIR_PADDLE,  COLOR_WHITE,   COLOR_BLACK);
		init_pair(CPAIR_CHARM,   COLOR_YELLOW,  COLOR_BLACK);
		init_pair(CPAIR_MISSILE, COLOR_WHITE,   COLOR_BLACK);
		init_pair(CPAIR_STATUS,  COLOR_BLACK,   COLOR_WHITE);
		init_pair(CPAIR_BRICK_U, COLOR_BLACK,   COLOR_RED);
		init_pair(CPAIR_BRICK_1, COLOR_GREEN,   brick_bg);
		init_pair(CPAIR_BRICK_2, COLOR_BLUE,    brick_bg);
		init_pair(CPAIR_BRICK_3, COLOR_YELLOW,  brick_bg);
		init_pair(CPAIR_BRICK_4, COLOR_MAGENTA, brick_bg);
		init_pair(CPAIR_BRICK_5, COLOR_CYAN,    brick_bg);
		init_pair(CPAIR_BRICK_6, COLOR_RED,     brick_bg);

		wattrset(g_stat_win, COLOR_PAIR(CPAIR_STATUS));
	}

	ptrarr_init(&g_balls);
	ptrarr_init(&g_bricks);
	ptrarr_init(&g_charms);
	ptrarr_init(&g_missiles);

	welcome_screen();

	clear_board();
	touchwin(g_game_win);
	wrefresh(g_game_win);
	halfdelay(1);
	keypad(g_game_win, TRUE);

	/* TODO: menu */

	do
	{
		new_game();
		paddle_show(&g_paddle);
		game_loop();
		paddle_hide(&g_paddle);
	}
	while(keep_going);




	endwin();
	return 0;
}
